@@ -9,16 +9,18 @@
 
 module Text.XML.HaXml.Html.Parse
   ( htmlParse
+  , htmlParse'
   ) where
 
 import Prelude hiding (either,maybe,sequence)
 import qualified Prelude (either)
-import Maybe hiding (maybe)
-import Char (toLower, isSpace, isDigit, isHexDigit)
+import Data.Maybe hiding (maybe)
+import Data.Char (toLower, {-isSpace,-} isDigit, isHexDigit)
 import Numeric (readDec,readHex)
-import Monad
+import Control.Monad
 
 import Text.XML.HaXml.Types
+import Text.XML.HaXml.Namespaces
 import Text.XML.HaXml.Lex
 import Text.XML.HaXml.Posn
 import Text.ParserCombinators.Poly.Plain
@@ -46,14 +48,14 @@ debug _ = return ()
 --   contents of the file.  The result is the generic representation of
 --   an XML document.  Any errors cause program failure with message to stderr.
 htmlParse :: String -> String -> Document Posn
-htmlParse name = Prelude.either error id . htmlParse' name
+htmlParse file = Prelude.either error id . htmlParse' file
 
 -- | The first argument is the name of the file, the second is the string
 --   contents of the file.  The result is the generic representation of
 --   an XML document.  Any parsing errors are returned in the @Either@ type.
 htmlParse' :: String -> String -> Either String (Document Posn)
-htmlParse' name = Prelude.either Left (Right . simplify) . fst
-                  . runParser document . xmlLex name
+htmlParse' file = Prelude.either Left (Right . simplify) . fst
+                  . runParser document . xmlLex file
 
 ---- Document simplification ----
 
@@ -61,9 +63,10 @@ simplify :: Document i -> Document i
 simplify (Document p st (Elem n avs cs) ms) =
     Document p st (Elem n avs (deepfilter simp cs)) ms
   where
-    simp (CElem (Elem "null" [] []) _) = False
-    simp (CElem (Elem  t     _  []) _) | t `elem` ["font","p","i","b","em"
-                                                  ,"tt","big","small"] = False
+    simp (CElem (Elem (N "null") [] []) _) = False
+    simp (CElem (Elem t _ []) _)
+        | localName t `elem` ["font","p","i","b","em","tt","big","small"]
+                                           = False
  -- simp (CString False s _) | all isSpace s = False
     simp _ = True
     deepfilter f =
@@ -77,7 +80,7 @@ selfclosingtags :: [String]
 selfclosingtags = ["img","hr","br","meta","col","link","base"
                   ,"param","area","frame","input"]
 
---closing this, implicitly closes any of those which are contained in it
+-- closing this, implicitly closes any of those which are contained in it
 closeInnerTags :: [(String,[String])]
 closeInnerTags =
   [ ("ul",      ["li"])
@@ -100,21 +103,21 @@ closeInnerTags =
   , ("body",    ["p"])
   ]
 
---opening this, implicitly closes that
+-- opening this, implicitly closes that
 closes :: Name -> Name -> Bool
-"a"  `closes` "a"   =  True
-"li" `closes` "li"  =  True
-"th" `closes`  t    | t `elem` ["th","td"]      =  True
-"td" `closes`  t    | t `elem` ["th","td"]      =  True
-"tr" `closes`  t    | t `elem` ["th","td","tr"] =  True
-"dt" `closes`  t    | t `elem` ["dt","dd"]      =  True
-"dd" `closes`  t    | t `elem` ["dt","dd"]      =  True
-"form"  `closes` "form"      = True
-"label" `closes` "label"     = True
-_       `closes` "option"    = True
-"thead" `closes` t  | t `elem` ["colgroup"]          = True
-"tfoot" `closes` t  | t `elem` ["thead","colgroup"]  = True
-"tbody" `closes` t  | t `elem` ["tbody","tfoot","thead","colgroup"] = True
+"a"     `closes` "a"      =  True
+"li"    `closes` "li"     =  True
+"th"    `closes`  t       | t `elem` ["th","td"]      =  True
+"td"    `closes`  t       | t `elem` ["th","td"]      =  True
+"tr"    `closes`  t       | t `elem` ["th","td","tr"] =  True
+"dt"    `closes`  t       | t `elem` ["dt","dd"]      =  True
+"dd"    `closes`  t       | t `elem` ["dt","dd"]      =  True
+"form"  `closes` "form"   = True
+"label" `closes` "label"  = True
+_       `closes` "option" = True
+"thead" `closes` t        | t `elem` ["colgroup"]          = True
+"tfoot" `closes` t        | t `elem` ["thead","colgroup"]  = True
+"tbody" `closes` t        | t `elem` ["tbody","tfoot","thead","colgroup"] = True
 "colgroup" `closes` "colgroup"  = True
 t `closes` "p"
     | t `elem` ["p","h1","h2","h3","h4","h5","h6"
@@ -132,6 +135,9 @@ tok t = do (p,t') <- next
                       _ | t'==t     -> return t
                         | otherwise -> report fail (show t) p t'
 
+qname :: HParser QName
+qname = fmap N name
+
 name :: HParser Name
 --name = do {(p,TokName s) <- next; return s}
 name = do (p,tok) <- next
@@ -197,7 +203,7 @@ document = do
     ms    <- many misc
     return (Document p emptyST (case map snd ht of
                                   [e] -> e
-                                  es  -> Elem "html" [] (map mkCElem es))
+                                  es  -> Elem (N "html") [] (map mkCElem es))
                                ms)
   where mkCElem e = CElem e noPos
 
@@ -263,7 +269,7 @@ doctypedecl = do
     tok TokSpecialOpen
     tok (TokSpecial DOCTYPEx)
     commit $ do
-      n <- name
+      n <- qname
       eid <- maybe externalid
 --    es <- maybe (bracket (tok TokSqOpen) (tok TokSqClose)) (many markupdecl)
       tok TokAnyClose  `onFail` failP "missing > in DOCTYPE decl"
@@ -314,39 +320,39 @@ element :: Name -> HParser (Stack,Element Posn)
 element ctx =
   do
     tok TokAnyOpen
-    (ElemTag e avs) <- elemtag
+    (ElemTag (N e) avs) <- elemtag
     ( if e `closes` ctx then
          -- insert the missing close-tag, fail forward, and reparse.
          ( do debug ("/")
               unparse ([TokEndOpen, TokName ctx, TokAnyClose,
                         TokAnyOpen, TokName e] ++ reformatAttrs avs)
-              return ([], Elem "null" [] []))
+              return ([], Elem (N "null") [] []))
       else if e `elem` selfclosingtags then
          -- complete the parse straightaway.
          ( do tok TokEndClose	-- self-closing <tag /> 
               debug (e++"[+]")
-              return ([], Elem e avs [])) `onFail`
+              return ([], Elem (N e) avs [])) `onFail`
      --  ( do tok TokAnyClose	-- sequence <tag></tag>	(**not HTML?**)
      --       debug (e++"[+")
-     --       n <- bracket (tok TokEndOpen) (tok TokAnyClose) name
+     --       n <- bracket (tok TokEndOpen) (tok TokAnyClose) qname
      --       debug "]"
      --       if e == (map toLower n :: Name) 
      --         then return ([], Elem e avs [])      
      --         else return (error "no nesting in empty tag")) `onFail`
          ( do tok TokAnyClose	-- <tag> with no close (e.g. <IMG>)
               debug (e++"[+]")
-              return ([], Elem e avs []))
+              return ([], Elem (N e) avs []))
       else
         (( do tok TokEndClose
               debug (e++"[]")
-              return ([], Elem e avs [])) `onFail`
+              return ([], Elem (N e) avs [])) `onFail`
          ( do tok TokAnyClose `onFail` failP "missing > or /> in element tag"
               debug (e++"[")
            -- zz <- many (content e)
-           -- n <- bracket (tok TokEndOpen) (tok TokAnyClose) name
+           -- n <- bracket (tok TokEndOpen) (tok TokAnyClose) qname
               zz <- manyFinally (content e)
                                 (tok TokEndOpen)
-              n <- name
+              (N n) <- qname
               commit (tok TokAnyClose)
               debug "]"
               let (ss,cs) = unzip zz
@@ -354,11 +360,11 @@ element ctx =
               ( if e == (map toLower n :: Name) then
                   do unparse (reformatTags (closeInner e s))
                      debug "^"
-                     return ([], Elem e avs cs)
+                     return ([], Elem (N e) avs cs)
                 else
                   do unparse [TokEndOpen, TokName n, TokAnyClose]
                      debug "-"
-                     return (((e,avs):s), Elem e avs cs))
+                     return (((e,avs):s), Elem (N e) avs cs))
          ) `onFail` failP ("failed to repair non-matching tags in context: "++ctx)))
 
 closeInner :: Name -> [(Name,[Attribute])] -> [(Name,[Attribute])]
@@ -371,12 +377,12 @@ unparse :: [TokenT] -> Parser (Posn, TokenT) ()
 unparse ts = do p <- posn
                 reparse (zip (repeat p) ts)
 
-reformatAttrs :: [(String, AttValue)] -> [TokenT]
+reformatAttrs :: [(QName, AttValue)] -> [TokenT]
 reformatAttrs avs = concatMap f0 avs
-    where f0 (a, v@(AttValue _)) = [TokName a, TokEqual, TokQuote,
-                                       TokFreeText (show v), TokQuote]
+    where f0 (a, v@(AttValue _)) = [ TokName (printableName a), TokEqual
+                                   , TokQuote, TokFreeText (show v), TokQuote ]
 
-reformatTags :: [(String, [(String, AttValue)])] -> [TokenT]
+reformatTags :: [(String, [(QName, AttValue)])] -> [TokenT]
 reformatTags ts = concatMap f0 ts
     where f0 (t,avs) = [TokAnyOpen, TokName t]++reformatAttrs avs++[TokAnyClose]
 
@@ -393,23 +399,23 @@ content ctx = do { p <- posn ; content' p }
 ----
 elemtag :: HParser ElemTag
 elemtag = do
-    n <- name `adjustErrBad` ("malformed element tag\n"++)
+    (N n) <- qname `adjustErrBad` ("malformed element tag\n"++)
     as <- many attribute
-    return (ElemTag (map toLower n) as)
+    return (ElemTag (N  $ map toLower n) as)
 
 attribute :: HParser Attribute
 attribute = do
-    n <- name
+    (N n) <- qname
     v <- (do tok TokEqual
              attvalue) `onFail`
          (return (AttValue [Left "TRUE"]))
-    return (map toLower n,v)
+    return (N $ map toLower n, v)
 
 --elementdecl :: HParser ElementDecl
 --elementdecl = do
 --    tok TokSpecialOpen
 --    tok (TokSpecial ELEMENTx)
---    n <- name `onFail` failP "missing identifier in ELEMENT decl"
+--    n <- qname `onFail` failP "missing identifier in ELEMENT decl"
 --    c <- contentspec `onFail` failP "missing content spec in ELEMENT decl"
 --    tok TokAnyClose `onFail` failP "expected > terminating ELEMENT decl"
 --    return (ElementDecl n c)
@@ -434,7 +440,7 @@ attribute = do
 --
 --cp :: HParser CP
 --cp =
---    ( do n <- name
+--    ( do n <- qname
 --         m <- modifier
 --         return (TagName n m)) `onFail`
 --    ( do ss <- sequence
@@ -461,7 +467,7 @@ attribute = do
 --  where
 --    cont = ( tok TokBraClose >> return PCDATA) `onFail`
 --           ( do cs <- many ( do tok TokPipe
---                                n <- name
+--                                n <- qname
 --                                return n)
 --                tok TokBraClose
 --                tok TokStar
@@ -471,14 +477,14 @@ attribute = do
 --attlistdecl = do
 --    tok TokSpecialOpen
 --    tok (TokSpecial ATTLISTx)
---    n <- name `onFail` failP "missing identifier in ATTLIST"
+--    n <- qname `onFail` failP "missing identifier in ATTLIST"
 --    ds <- many attdef
 --    tok TokAnyClose `onFail` failP "missing > terminating ATTLIST"
 --    return (AttListDecl n ds)
 --
 --attdef :: HParser AttDef
 --attdef = do
---    n <- name
+--    n <- qname
 --    t <- atttype `onFail` failP "missing attribute type in attlist defn"
 --    d <- defaultdecl
 --    return (AttDef n t d)
@@ -559,7 +565,7 @@ reference = do
                     = return . RefChar . fst . head . readHex $ i
     val ('#':i)     | all isDigit i
                     = return . RefChar . fst . head . readDec $ i
-    val name        = return . RefEntity $ name
+    val ent         = return . RefEntity $ ent
 
 {-
 reference :: HParser Reference
