@@ -8,9 +8,10 @@ module Text.XML.HaXml.Wrappers
 
 -- imports required for processXmlWith and fix2Args
 import Prelude hiding (filter)
-import System
-import IO
-import List (isSuffixOf)
+import System.Exit
+import System.Environment
+import System.IO
+import Data.List  (isSuffixOf)
 import Control.Monad (when)
 
 import Text.XML.HaXml.Types       (Document(..),Content(..))
@@ -30,7 +31,7 @@ fix2Args :: IO (String,String)
 fix2Args = do
   args <- getArgs
   when ("--version" `elem` args) $ do
-      putStrLn $ "part of HaXml-"++MYVERSION
+      putStrLn $ "part of HaXml-"++show MAJOR.MINOR
       exitWith ExitSuccess
   when ("--help" `elem` args) $ do
       putStrLn $ "See http://haskell.org/HaXml"
@@ -68,8 +69,9 @@ processXmlWith f = do
     onContent file filter (Document p s e m) =
         case filter (CElem e (posInNewCxt file Nothing)) of
             [CElem e' _] -> Document p s e' m
-            []           -> error "produced no output"
-            _            -> error "produced more than one output"
+            []           -> error $ "filtering"++file++"produced no output"
+            _            -> error $ "filtering"++file++
+                                    "produced more than one output document"
 
 -- | The wrapper @onContent@ simply applies a given content filter to a
 --   document.  Ambiguous or empty results raise an error exception.
